/*
 * Copyright (C) 2019-2024 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *		Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *		Author: Mohd Husaam Mehdi <husaam.mehdi@iopsys.eu>
 */

#include "dns_sd.h"
#include <json-c/json.h>
#include <libgen.h>

/*************************************************************
* ENTRY METHOD
**************************************************************/
static int browseDNSSDAdvertiseTextRecordInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dm_data *curr_data = NULL;
	LIST_HEAD(dup_list);
	char *inst = NULL;

	struct uci_section *parent_data = ((struct dm_data *)prev_data)->dmmap_section;

	char *filename;
	dmuci_get_value_by_section_string(parent_data, "file", &filename);

	struct uci_section *s = NULL;
	uci_path_foreach_sections(bbfdm, "dmmap_umdns", "text_record", s) {
		char *p_name = NULL;
		dmuci_get_value_by_section_string(s, "parent", &p_name);
		if (strcmp(p_name, filename) == 0) {
			add_dmmap_config_dup_list(&dup_list, NULL, s);
		}
	}

	list_for_each_entry(curr_data, &dup_list, list) {
		inst = handle_instance(dmctx, parent_node, curr_data->dmmap_section, "txt_instance", "txt_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, curr_data, inst) == DM_STOP)
			break;
	}

	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int get_dns_sd_advertise_text_record_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseDNSSDAdvertiseTextRecordInst);
	dmasprintf(value, "%d", cnt);

	return 0;
}

static int get_dns_sd_advertise_text_record_key(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "key", value);
	return 0;
}

static int set_dns_sd_advertise_text_record_key(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 255, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			char *parent = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "parent", &parent);

			if ((parent == NULL) || (strcmp(parent, "") == 0))
				return FAULT_9002;

			json_object *jobj = json_object_from_file(parent);
			if (!jobj)
				return FAULT_9002;

			char *old_key = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "key", &old_key);

			// old record is used for updating existing value of record
			char old_record[256] = {};
			if ((old_key != NULL) && (strcmp(old_key, "") != 0))
				snprintf(old_record, sizeof(old_record), "%s=", old_key);
			else
				snprintf(old_record, sizeof(old_record), "=");

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "key", value);

			char new_record[256] = {};
			snprintf(new_record, sizeof(new_record), "%s=", value);

			char *key_value = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "key_value", &key_value);
			if ((key_value != NULL) || (strcmp(key_value, "") != 0)) {
				strncat(new_record, key_value, sizeof(new_record) - strlen(new_record) - 1);
				strncat(old_record, key_value, sizeof(old_record) - strlen(old_record) - 1);
			}

			json_object *jval = NULL;
			json_object *option = NULL;
			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}

				json_object_object_get_ex(val, "txt", &option);
				if (option == NULL) {
					jval = json_object_new_array();
					json_object_array_add(jval, json_object_new_string(new_record));
					json_object_object_add(val, "txt", jval);
				} else {
					int record_found = 0;
					for (int i = 0; i < json_object_array_length(option); i++) {
						jval = json_object_array_get_idx(option, i);
						if (strcmp(json_object_get_string(jval), old_record) == 0) {
							json_object_set_string(jval, new_record);
							record_found = 1;
							break;
						}
					}

					if (record_found == 0)
						json_object_array_add(option, json_object_new_string(new_record));
				}
			}

			json_object_to_file(parent, jobj);
			json_object_put(jobj);
			// call ubus else written param will not take effect
			json_object *dummy = NULL;
			dmubus_call("umdns", "reload", UBUS_ARGS{0}, 0, &dummy);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_text_record_value(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "key_value", value);

	return 0;
}

static int set_dns_sd_advertise_text_record_value(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 255, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			char *parent = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "parent", &parent);

			if ((parent == NULL) || (strcmp(parent, "") == 0))
				return FAULT_9002;

			json_object *jobj = json_object_from_file(parent);
			if (!jobj)
				return FAULT_9002;

			char old_record[256] = "", new_record[256] = "";
			char *param_key = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "key", &param_key);
			if ((param_key != NULL) && (strcmp(param_key, "") != 0)) {
				snprintf(old_record, sizeof(old_record), "%s=", param_key);
				snprintf(new_record, sizeof(new_record), "%s=", param_key);
			} else {
				snprintf(old_record, sizeof(old_record), "=");
				snprintf(new_record, sizeof(new_record), "=");
			}

			char *old_key_value = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "key_value", &old_key_value);
			if ((old_key_value != NULL) && (strcmp(old_key_value, "") != 0))
				strncat(old_record, old_key_value, sizeof(old_record) - strlen(old_record) - 1);

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "key_value", value);
			strncat(new_record, value, sizeof(new_record) - strlen(new_record) - 1);

			json_object *jval = NULL;
			json_object *option = NULL;
			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}

				json_object_object_get_ex(val, "txt", &option);
				if (option == NULL) {
					jval = json_object_new_array();
					json_object_array_add(jval, json_object_new_string(new_record));
					json_object_object_add(val, "txt", jval);
				} else {
					int record_found = 0;
					for (int i = 0; i < json_object_array_length(option); i++) {
						jval = json_object_array_get_idx(option, i);
						if (strcmp(json_object_get_string(jval), old_record) == 0) {
							json_object_set_string(jval, new_record);
							record_found = 1;
							break;
						}
					}

					if (record_found == 0) {
						json_object_array_add(option, json_object_new_string(new_record));
					}
				}
			}

			json_object_to_file(parent, jobj);
			json_object_put(jobj);
			// call ubus else written param will not take effect
			json_object *dummy = NULL;
			dmubus_call("umdns", "reload", UBUS_ARGS{0}, 0, &dummy);
			break;
	}

	return 0;
}

static int browseDNSSDServiceInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dm_data curr_data = {0};
	struct srv_info current_srv_info_obj = {0};

	json_object *ubus_result_browse = NULL, *service_objects = NULL;
	int id = 0;
	char *inst = NULL;

	// call ubus
	dmubus_call("umdns", "browse", UBUS_ARGS{0}, 0, &ubus_result_browse);

	// return from here if result empty
	DM_ASSERT(ubus_result_browse, bbfdm_set_fault_message(dmctx, "ubus call umdns browse did not work"));

	json_object_object_get_ex(ubus_result_browse, "discovered_services", &service_objects);
	json_object_object_foreach(service_objects, key, val) {
		char *proto_position = NULL;
		char *str = NULL;

		// example of key: "_http._tcp"
		current_srv_info_obj.application_protocol = dmstrdup(key);

		// for readability
		str = current_srv_info_obj.application_protocol;

		// find the point where transport protocol starts
		if (DM_STRSTR(str, "._tcp")) {
			proto_position = DM_STRSTR(str, "._tcp");
			current_srv_info_obj.transport_protocol_flag = TRANSPORT_TCP;
		} else if (DM_STRSTR(str, "._udp")) {
			proto_position = DM_STRSTR(str, "._udp");
			current_srv_info_obj.transport_protocol_flag = TRANSPORT_UDP;
		} else {
			continue;
		}

		// mark the end of application protocol
		*proto_position = '\0';

		// find and remove leading underscore
		proto_position = DM_STRSTR(str, "_");
		if (proto_position)
			current_srv_info_obj.application_protocol = ++proto_position;

		json_object_object_foreach(val, inner_key, inner_val) {
			// update our object
			current_srv_info_obj.instance_name = dmstrdup(inner_key);

			curr_data.json_object = inner_val;
			curr_data.additional_data = (void *)(&current_srv_info_obj);

			inst = handle_instance_without_section(dmctx, parent_node, ++id);

			if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)&curr_data, inst) == DM_STOP)
				break;
		}
	}

	return 0;
}

static int browseDNSSDServiceTextRecordInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dm_data curr_data = {0};

	char *inst = NULL;
	char *key_value_str = NULL;
	char *dup_key_value_str = NULL;
	char *pch = NULL, *spch = NULL;

	key_value_str = dmjson_get_value_array_all(((struct dm_data *)prev_data)->json_object, ",", 1, "txt");

	if (key_value_str) {
		int id = 0;
		dup_key_value_str = dmstrdup(key_value_str);
		for (pch = strtok_r(dup_key_value_str, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			inst = handle_instance_without_section(dmctx, parent_node, ++id);

			curr_data.additional_data = (void *)pch;
			if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)&curr_data, inst) == DM_STOP)
				break;
		}
	}

	return 0;
}

/*************************************************************
* GET & SET PARAM
**************************************************************/
static int get_dns_sd_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_option_value_fallback_def("umdns", "@umdns[0]", "enable", "1");
	return 0;
}

static int set_dns_sd_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value("umdns", "@umdns[0]", "enable", value);
			break;
	}

	return 0;
}

static int get_dns_sd_service_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseDNSSDServiceInst);
	dmasprintf(value, "%d", cnt);

	return 0;
}

static int get_dns_sd_service_instance_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *instance_name;

	instance_name = dmstrdup(((struct srv_info *)((struct dm_data *)data)->additional_data)->instance_name);
	if (DM_STRLEN(instance_name) > 64 ) {
		instance_name[64] = '\0';
	}

	*value = instance_name;
	return 0;
}

static int get_dns_sd_service_application_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *proto;

	proto = dmstrdup(((struct srv_info *)((struct dm_data *)data)->additional_data)->application_protocol);
	if (DM_STRLEN(proto) > 16 ) {
		proto[16] = '\0';
	}

	*value = proto;
	return 0;
}

static int get_dns_sd_service_transport_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (((struct srv_info *)((struct dm_data *)data)->additional_data)->transport_protocol_flag == TRANSPORT_TCP)
		*value = "TCP";
	else
		*value = "UDP";

	return 0;
}

static int get_dns_sd_service_domain(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *domain;

	domain = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "domain"));
	if (DM_STRLEN(domain) > 256 ) {
		domain[256] = '\0';
	}

	*value = domain;
	return 0;
}

static int get_dns_sd_service_port(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *port;

	port = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "port"));

	*value = port;
	return 0;
}

static int get_dns_sd_service_target(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *target;

	target = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "host"));

	*value = target;
	return 0;
}

static int get_dns_sd_service_status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	// inactive services are not present in output of ubus
	*value = "LeaseActive";
	return 0;
}

static int get_dns_sd_service_last_update(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *last_updated;

	last_updated = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "last_update"));

	*value = last_updated;
	return 0;
}

static int get_dns_sd_service_host(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *ipv4_address = NULL;
	json_object *ipv4_array = NULL;
	unsigned idx = 0;
	char hosts_list[HOSTS_LIST_SIZE] = {0};

	dmjson_foreach_value_in_array(((struct dm_data *)data)->json_object, ipv4_array, ipv4_address, idx, 1, "ipv4") {
		bbfdm_get_references(ctx, MATCH_FIRST, "Device.Hosts.Host.", "IPAddress", ipv4_address, hosts_list, sizeof(hosts_list));
	}

	*value = dmstrdup(hosts_list);
	return 0;
}

static int get_dns_sd_service_time_to_live(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *ttl;

	ttl = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "ttl"));

	*value = ttl;
	return 0;
}

static int get_dns_sd_service_priority(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *priority;

	priority = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "priority"));

	*value = priority;
	return 0;
}

static int get_dns_sd_service_weight(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *weight;

	weight = dmstrdup(dmjson_get_value(((struct dm_data *)data)->json_object, 1, "weight"));

	*value = weight;
	return 0;
}

static int get_dns_sd_service_text_record_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseDNSSDServiceTextRecordInst);
	dmasprintf(value, "%d", cnt);

	return 0;
}

static int get_dns_sd_service_text_record_key(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *key_value_str = dmstrdup((char *)((struct dm_data *)data)->additional_data);

	char *key_end_position = DM_STRSTR(key_value_str, "=");

	// mark end of key
	if (key_end_position)
		*key_end_position = '\0';

	*value = key_value_str;
	return 0;
}

static int get_dns_sd_service_text_record_value(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *key_value_str = dmstrdup((char *)((struct dm_data *)data)->additional_data);

	char *value_start_position = DM_STRSTR(key_value_str, "=");

	// point to start of value
	if (++value_start_position)
		key_value_str = value_start_position;

	*value = key_value_str;
	return 0;
}

static int addObjDNSSDAdvertiseTextRecord(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *parent_data = ((struct dm_data *)data)->dmmap_section;
	char *filename = NULL;
	dmuci_get_value_by_section_string(parent_data, "file", &filename);
	if (filename == NULL)
		return 0;

	struct uci_section *dmmap_s = NULL;
	dmuci_add_section_bbfdm("dmmap_umdns", "text_record", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "parent", filename);
	dmuci_set_value_by_section(dmmap_s, "bbf", "1");

	return 0;
}

static int delObjDNSSDAdvertiseTextRecord(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	switch (del_action) {
		case DEL_INST:
			struct dm_data *p = (struct dm_data *)data;
			char *parent = NULL;

			dmuci_get_value_by_section_string(p->dmmap_section, "parent", &parent);

			json_object *jobj = json_object_from_file(parent);
			if (!jobj)
				return FAULT_9002;

			char *param_key = NULL, *key_value = NULL;
			dmuci_get_value_by_section_string(p->dmmap_section, "key", &param_key);
			dmuci_get_value_by_section_string(p->dmmap_section, "key_value", &key_value);
			char record[512] = "";
			snprintf(record, sizeof(record), "%s=%s", param_key, key_value);
			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}

				json_object *jval = NULL;
				json_object *option = NULL;
				json_object_object_get_ex(val, "txt", &option);
				for (int i = 0; i < json_object_array_length(option); i++) {
					jval = json_object_array_get_idx(option, i);
					if (strcmp(json_object_get_string(jval), record) == 0) {
						json_object_array_del_idx(option, i, 1);
						break;
					}
				}
				json_object_to_file(parent, jobj);
			}
			json_object_put(jobj);
			dmuci_delete_by_section(p->dmmap_section, NULL, NULL);
			break;
		case DEL_ALL:
			break;
	}

	return 0;
}

static int addObjDNSSDAdvertise(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	char file_name[256] = {};
	snprintf(file_name, sizeof(file_name), "/etc/umdns/iowrt_%s.json", *instance);

	json_object *jobj = json_object_new_object();
	json_object *inner_obj = json_object_new_object();
	char obj_name[256] = {};
	snprintf(obj_name, sizeof(obj_name), "iowrt_%s", *instance);
	json_object_object_add(jobj, obj_name, inner_obj);
	if (json_object_to_file(file_name, jobj) != 0)
		return FAULT_9002;

	json_object_put(inner_obj);
	json_object_put(jobj);
	// add new section
	struct uci_section *dmmap_s = NULL;
	dmuci_add_section_bbfdm("dmmap_umdns", "advertise", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "enable", "0");
	dmuci_set_value_by_section(dmmap_s, "auto", "0");
	dmuci_set_value_by_section(dmmap_s, "application_protocol", "");
	dmuci_set_value_by_section(dmmap_s, "transport_protocol", "tcp");
	dmuci_set_value_by_section(dmmap_s, "domain", "local");
	dmuci_set_value_by_section(dmmap_s, "port", "");
	dmuci_set_value_by_section(dmmap_s, "file", file_name);
	dmuci_set_value_by_section(dmmap_s, "bbf", "1");
	return 0;
}

static int delObjDNSSDAdvertise(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct dm_data *p = (struct dm_data *)data;
	char *file_name = NULL;
	switch (del_action) {
		case DEL_INST:
			dmuci_get_value_by_section_string(p->dmmap_section, "file", &file_name);

			// delete all child text objects
			struct uci_section *s = NULL;
			uci_path_foreach_sections(bbfdm, "dmmap_umdns", "text_record", s) {
				char *parent_name = NULL;
				dmuci_get_value_by_section_string(s, "parent", &parent_name);
				if (strcmp(parent_name, file_name) == 0) {
					dmuci_delete_by_section(s, NULL, NULL);
				}
			}

			// now delete the advertise section
			dmuci_delete_by_section(p->dmmap_section, NULL, NULL);

			// finally delete the json file corresponding to advertise
			if (remove(file_name) != 0)
				return FAULT_9002;
			break;
		case DEL_ALL:
			break;
	}

	return 0;
}

static int browseDNSSDAdvertiseInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	char *inst = NULL;
	struct dm_data *curr_data = NULL;
	LIST_HEAD(dup_list);

	DIR *dir = opendir("/etc/umdns/");

	if (dir == NULL) {
		return 0;
	}

	struct dirent *entry;
	while ((entry = readdir(dir)) != NULL) {
		if (entry->d_type != DT_REG)
			continue;

		// Process only JSON files
		const char *ext = strrchr(entry->d_name, '.');
		if (ext && strcmp(ext, ".json") != 0)
			continue;

		char filepath[1024];
		snprintf(filepath, sizeof(filepath), "/etc/umdns/%s", entry->d_name);
		// check if section already exists for this file in dmmap, if not,
		// create one
		bool instance_found = false;
		struct uci_section *s = NULL;
		uci_path_foreach_sections(bbfdm, "dmmap_umdns", "advertise", s) {
			char *file_name = NULL;
			dmuci_get_value_by_section_string(s, "file", &file_name);
			if (strcmp(file_name, filepath) == 0) {
				instance_found = true;
				break;
			}
		}

		struct uci_section *dmmap_s = NULL, *dmmap_sect = NULL;
		if (!instance_found) {
			// create dmmap entry for this
			dmuci_add_section_bbfdm("dmmap_umdns", "advertise", &dmmap_s);
			dmuci_set_value_by_section(dmmap_s, "file", filepath);
			dmuci_set_value_by_section(dmmap_s, "enable", "1");
			add_dmmap_config_dup_list(&dup_list, NULL, dmmap_s);
			dmmap_sect = dmmap_s;
		} else {
			add_dmmap_config_dup_list(&dup_list, NULL, s);
			dmmap_sect = s;
			char *creator_bbf = NULL;
			dmuci_get_value_by_section_string(s, "bbf", &creator_bbf);
			if (strcmp(creator_bbf, "1") == 0)
				continue;
		}

		json_object *jobj = json_object_from_file(filepath);
		if (jobj == NULL)
			continue;

		json_object *holder = NULL;
		json_object_object_foreach(jobj, key, val) {
			if (key == NULL)
				continue;

			char *inst_name = NULL;
			dmuci_get_value_by_section_string(dmmap_sect, "instance_name", &inst_name);

			if (DM_STRLEN(inst_name) == 0)
				dmuci_set_value_by_section(dmmap_sect, "instance_name", key);

			if (val == NULL)
				continue;

			json_object_object_get_ex(val, "service", &holder);
			if (holder != NULL) {
				char *service = dmstrdup(json_object_get_string(holder));

				if ((service != NULL) && strcmp(service, "") != 0) {
					char *token = strtok(service, ".");
					if (token && token[0] == '_')
						dmuci_set_value_by_section(dmmap_sect, "application_protocol", token + 1);

					token = strtok(NULL, ".");
					if (token && token[0] == '_')
						dmuci_set_value_by_section(dmmap_sect, "transport_protocol", token + 1);

					token = strtok(NULL, ".");
					if (token)
						dmuci_set_value_by_section(dmmap_sect, "domain", token);
				}
			}

			json_object_object_get_ex(val, "port", &holder);

			if (holder != NULL) {
				char *port = dmstrdup(json_object_get_string(holder));

				if (port)
					dmuci_set_value_by_section(dmmap_sect, "port", port);
			}

			// now parse and generate the txt section
			// first clear all txt section for this filename
			struct uci_section *txt_s = NULL;
			uci_path_foreach_sections(bbfdm, "dmmap_umdns", "text_record", txt_s) {
				char *creator = NULL;
				dmuci_get_value_by_section_string(txt_s, "bbf", &creator);
				if (strcmp(creator, "1") == 0)
					continue;

				char *f_name = NULL;
				dmuci_get_value_by_section_string(txt_s, "parent", &f_name);
				if (strcmp(f_name, filepath) == 0)
					dmuci_delete_by_section(txt_s, NULL, NULL);
			}

			json_object_object_get_ex(val, "txt", &holder);

			if (holder != NULL) {
				int length = json_object_array_length(holder);
				int i;
				for (i = 0; i < length; i++) {
					json_object *txt_element = json_object_array_get_idx(holder, i);
					char *txt_value = dmstrdup(json_object_get_string(txt_element));
					char param_key[256] = "", key_value[256] = "";

					char *token = strtok(txt_value, "=");
					if (token != NULL) {
						snprintf(param_key, sizeof(param_key), "%s", token);
						token = strtok(NULL, "");
						if (token != NULL)
							snprintf(key_value, sizeof(key_value), "%s", token);
					}

					// now create uci section for this txt record
					dmuci_add_section_bbfdm("dmmap_umdns", "text_record", &txt_s);
					dmuci_set_value_by_section(txt_s, "parent", filepath);

					if (strcmp(param_key, "") != 0)
						dmuci_set_value_by_section(txt_s, "key", param_key);

					if (strcmp(key_value, "") != 0)
						dmuci_set_value_by_section(txt_s, "key_value", key_value);

					json_object_put(txt_element);
				}
			}
		}

		json_object_put(jobj);
	}

	list_for_each_entry(curr_data, &dup_list, list) {
		inst = handle_instance(dmctx, parent_node, curr_data->dmmap_section, "adv_instance", "adv_alias");
		if (DM_LINK_INST_OBJ(dmctx, parent_node, curr_data, inst) == DM_STOP)
			break;
	}

	free_dmmap_config_dup_list(&dup_list);
	closedir(dir);
	return 0;
}

static int get_dns_sd_advertise_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseDNSSDAdvertiseInst);
	dmasprintf(value, "%d", cnt);

	return 0;
}

static int get_dns_sd_advertise_instance_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "instance_name", value);
	return 0;
}

static int set_dns_sd_advertise_instance_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 63, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "instance_name", value);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_application_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "application_protocol", value);
	return 0;
}

static int set_dns_sd_advertise_application_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 15, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			// first check if the json file for this service exists
			char *adv_file = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "file", &adv_file);
			if (strcmp(adv_file, "") == 0)
				return FAULT_9002;

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "application_protocol", value);
			json_object *jobj = json_object_from_file(adv_file);
			if (!jobj)
				return FAULT_9002;


			char new_service[256] = "";
			snprintf(new_service, sizeof(new_service), "_%s", value);

			char *transport_protocol = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "transport_protocol", &transport_protocol);
			if ((transport_protocol != NULL) && strcmp(transport_protocol, "") != 0) {
				strncat(new_service, "._", sizeof(new_service) - strlen(new_service) - 1);
				if (strcmp(transport_protocol, "TCP") == 0)
					strncat(new_service, "tcp", sizeof(new_service) - strlen(new_service) - 1);
				else
					strncat(new_service, "udp", sizeof(new_service) - strlen(new_service) - 1);
			}

			char *domain = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "domain", &domain);
			if ((domain != NULL) && strcmp(domain, "") != 0) {
				strncat(new_service, ".", sizeof(new_service) - strlen(new_service) - 1);
				strncat(new_service, domain, sizeof(new_service) - strlen(new_service) - 1);
			}

			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}

				json_object *option = NULL;
				json_object_object_get_ex(val, "service", &option);
				if (option == NULL) {
					json_object_object_add(val, "service", json_object_new_string(new_service));
				} else {
					json_object_set_string(option, new_service);
				}
				json_object_to_file(adv_file, jobj);
			}

			json_object_put(jobj);
			// call ubus else written param will not take effect
			json_object *dummy = NULL;
			dmubus_call("umdns", "reload", UBUS_ARGS{0}, 0, &dummy);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_transport_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *proto = NULL;
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "transport_protocol", &proto);

	if (strcmp(proto, "tcp") == 0)
		*value = "TCP";
	else if (strcmp(proto, "udp") == 0)
		*value = "UDP";
	else
		*value = "";

	return 0;
}


static int set_dns_sd_advertise_transport_protocol(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *adv_trans_proto[] = {"TCP", "UDP", NULL};

	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, adv_trans_proto, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			// first check if the json file for this service exists
			char *adv_file = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "file", &adv_file);
			if (strcmp(adv_file, "") == 0)
				return 0; // do nothing, just return

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "transport_protocol", value);
			json_object *jobj = json_object_from_file(adv_file);
			if (!jobj)
				return FAULT_9002;

			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}

				char new_service[256] = "";
				char *app_proto = NULL;
				dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "application_protocol", &app_proto);
				if (app_proto != NULL)
					snprintf(new_service, sizeof(new_service), "%s", app_proto);

				strncat(new_service, "._", sizeof(new_service) - strlen(new_service) - 1);
				if (strcmp(value, "TCP") == 0)
					strncat(new_service, "tcp", sizeof(new_service) - strlen(new_service) - 1);
				else
					strncat(new_service, "udp", sizeof(new_service) - strlen(new_service) - 1);

				char *domain = NULL;
				dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "domain", &domain);
				if ((domain != NULL) && strcmp(domain, "") != 0) {
					strncat(new_service, ".", sizeof(new_service) - strlen(new_service) - 1);
					strncat(new_service, domain, sizeof(new_service) - strlen(new_service) - 1);
				}

				json_object *option = NULL;
				json_object_object_get_ex(val, "service", &option);
				if (option == NULL) {
					json_object_object_add(val, "service", json_object_new_string(new_service));
				} else {
					json_object_set_string(option, new_service);
				}
				json_object_to_file(adv_file, jobj);
				json_object_put(option);
			}

			json_object_put(jobj);
			// call ubus else written param will not take effect
			json_object *dummy = NULL;
			dmubus_call("umdns", "reload", UBUS_ARGS{0}, 0, &dummy);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_domain(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "domain", value);
	return 0;
}

static int set_dns_sd_advertise_domain(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 256, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			// first check if the json file for this service exists
			char *adv_file = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "file", &adv_file);
			if (strcmp(adv_file, "") == 0)
				return 0; // do nothing, just return

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "domain", value);
			json_object *jobj = json_object_from_file(adv_file);
			if (!jobj)
				return FAULT_9002;

			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}

				char new_service[256] = "";
				char *app_proto = NULL;
				dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "application_protocol", &app_proto);
				if (app_proto != NULL)
					snprintf(new_service, sizeof(new_service), "%s", app_proto);

				char *t_proto = NULL;
				dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "transport_protocol", &t_proto);
				if ((t_proto != NULL) && strcmp(t_proto, "") != 0) {
					strncat(new_service, "._", sizeof(new_service) - strlen(new_service) - 1);
					if (strcmp(t_proto, "TCP") == 0)
						strncat(new_service, "tcp", sizeof(new_service) - strlen(new_service) - 1);
					else
						strncat(new_service, "udp", sizeof(new_service) - strlen(new_service) - 1);
				}

				strncat(new_service, ".", sizeof(new_service) - strlen(new_service) - 1);
				strncat(new_service, value, sizeof(new_service) - strlen(new_service) - 1);

				json_object *option = NULL;
				json_object_object_get_ex(val, "service", &option);
				if (option == NULL) {
					json_object_object_add(val, "service", json_object_new_string(new_service));
				} else {
					json_object_set_string(option, new_service);
				}
				json_object_to_file(adv_file, jobj);
				json_object_put(option);
			}

			json_object_put(jobj);
			// call ubus else written param will not take effect
			json_object *dummy = NULL;
			dmubus_call("umdns", "reload", UBUS_ARGS{0}, 0, &dummy);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_port(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "port", value);
	return 0;
}

static int set_dns_sd_advertise_port(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{"0","65535"}}, 1))
				return FAULT_9007;
			break;
		case VALUESET:
			// first check if the json file for this service exists
			char *adv_file = NULL;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "file", &adv_file);
			if (strcmp(adv_file, "") == 0)
				return 0; // do nothing, just return

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "port", value);
			json_object *jobj = json_object_from_file(adv_file);
			if (!jobj)
				return FAULT_9002;

			json_object_object_foreach(jobj, key, val) {
				if (key == NULL) {
					json_object_put(jobj);
					return FAULT_9002;
				}


				char *endptr;
				long int num = strtol(value, &endptr, 10);
				if (*endptr != '\0') {
					json_object_put(jobj);
					return FAULT_9002;
				}

				json_object *option = NULL;
				json_object_object_get_ex(val, "port", &option);
				if (option == NULL) {
					json_object_object_add(val, "port", json_object_new_int(num));
				} else {
					json_object_set_int(option, num);
				}

				json_object_to_file(adv_file, jobj);
				json_object_put(option);
			}
			json_object_put(jobj);
			// call ubus else written param will not take effect
			json_object *dummy = NULL;
			dmubus_call("umdns", "reload", UBUS_ARGS{0}, 0, &dummy);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dm_data *)data)->dmmap_section, "adv_alias", instance, value);
}

static int set_dns_sd_advertise_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dm_data *)data)->dmmap_section, "adv_alias", instance, value);
}

static int get_dns_sd_advertise_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "enable", value);
	return 0;
}

static int set_dns_sd_advertise_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:

			char *fname;
			dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "file", &fname);
			if (strcmp(fname, "") == 0)
				return 0;

			char destination_path[256];
			snprintf(destination_path, sizeof(destination_path), "/etc/umdns/tmp/%s", basename(fname));
			// Check if the destination directory exists
			struct stat st;
			if (stat(destination_path, &st) != 0) {
				// Create the directory if it doesn't exist
				if (mkdir(destination_path, 0755) != 0) {
					return FAULT_9002;
				}
			}

			dmuci_set_value_by_section(((struct dm_data *)data)->dmmap_section, "enable", value);
			FILE *file = fopen(fname, "r");
			if (file && (strcmp(value, "0") == 0)) {
				// move file to tmp location
				if(rename(fname, destination_path) != 0) {
					fclose(file);
					return FAULT_9002;
				}
			} else if (!file && (strcmp(value, "1") == 0)) {
				if(rename(destination_path, fname) != 0)
					return FAULT_9002;
			}

			if (file)
				fclose(file);
			break;
	}

	return 0;
}

static int get_dns_sd_advertise_status(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	// inactive services are not present in output of ubus
	char *enabled = NULL;
	dmuci_get_value_by_section_string(((struct dm_data *)data)->dmmap_section, "enable", &enabled);

	if (strcmp(enabled, "1") == 0)
		*value = "Enabled";
	else
		*value = "Disabled";

	return 0;
}

static int get_dns_sd_advertise_interface(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *net_list = NULL;

	dmuci_get_option_value_list("umdns", "@umdns[0]", "network", &net_list);

	if (net_list != NULL) {
		struct uci_element *e = NULL;
		char iface[256] = {0};

		uci_foreach_element(net_list, e) {
			bbfdm_get_references(ctx, MATCH_FIRST, "Device.IP.Interface.", "Name", e->name, iface, sizeof(iface));
		}

		*value = dmstrdup(iface);
		return 0;
	}

	*value = "";
	return 0;
}

static int set_dns_sd_advertise_interface(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_string_list(ctx, value, -1, -1, 1024, -1, -1, NULL, NULL))
				return FAULT_9007;

			if (*value != '\0')
				return FAULT_9007;

			break;
		case VALUESET:
			break;
	}
	return 0;
}

/**********************************************************************************************************************************
*                                            OBJ & LEAF DEFINITION
***********************************************************************************************************************************/
/* *** Device.DNS.SD. *** */
DMOBJ tDNSPluginObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"SD", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tDNSSDObj, tDNSSDParams, NULL, BBFDM_BOTH, NULL},
{0}
};

/* *** Device.DNS.SD. *** */
DMOBJ tDNSSDObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"Service", &DMREAD, NULL, NULL, NULL, browseDNSSDServiceInst, NULL, NULL, tDNSSDServiceObj, tDNSSDServiceParams, NULL, BBFDM_BOTH, NULL},
{"Advertise", &DMWRITE, addObjDNSSDAdvertise, delObjDNSSDAdvertise, NULL, browseDNSSDAdvertiseInst, NULL, NULL, tDNSSDAdvertiseObj, tDNSSDAdvertiseParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tDNSSDAdvertiseParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_dns_sd_advertise_enable, set_dns_sd_advertise_enable, BBFDM_BOTH},
{"Alias", &DMWRITE, DMT_STRING, get_dns_sd_advertise_alias, set_dns_sd_advertise_alias, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_dns_sd_advertise_status, NULL, BBFDM_BOTH},
{"Interface", &DMWRITE, DMT_STRING, get_dns_sd_advertise_interface, set_dns_sd_advertise_interface, BBFDM_BOTH, DM_FLAG_REFERENCE},
{"InstanceName", &DMWRITE, DMT_STRING, get_dns_sd_advertise_instance_name, set_dns_sd_advertise_instance_name, BBFDM_BOTH},
{"ApplicationProtocol", &DMWRITE, DMT_STRING, get_dns_sd_advertise_application_protocol, set_dns_sd_advertise_application_protocol, BBFDM_BOTH},
{"TransportProtocol", &DMWRITE, DMT_STRING, get_dns_sd_advertise_transport_protocol, set_dns_sd_advertise_transport_protocol, BBFDM_BOTH},
{"Domain", &DMWRITE, DMT_STRING, get_dns_sd_advertise_domain, set_dns_sd_advertise_domain, BBFDM_BOTH},
{"Port", &DMWRITE, DMT_UNINT, get_dns_sd_advertise_port, set_dns_sd_advertise_port, BBFDM_BOTH},
{"TextRecordNumberOfEntries", &DMREAD, DMT_UNINT, get_dns_sd_advertise_text_record_number_of_entries, NULL, BBFDM_BOTH},
{0}
};

DMOBJ tDNSSDAdvertiseObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"TextRecord", &DMWRITE, addObjDNSSDAdvertiseTextRecord, delObjDNSSDAdvertiseTextRecord, NULL, browseDNSSDAdvertiseTextRecordInst, NULL, NULL, NULL, tDNSSDAdvertiseTextRecordParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tDNSSDAdvertiseTextRecordParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Key", &DMWRITE, DMT_STRING, get_dns_sd_advertise_text_record_key, set_dns_sd_advertise_text_record_key, BBFDM_BOTH},
{"Value", &DMWRITE, DMT_STRING, get_dns_sd_advertise_text_record_value, set_dns_sd_advertise_text_record_value, BBFDM_BOTH},
{0}
};

DMLEAF tDNSSDParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Enable", &DMWRITE, DMT_BOOL, get_dns_sd_enable, set_dns_sd_enable, BBFDM_BOTH},
{"ServiceNumberOfEntries", &DMREAD, DMT_UNINT, get_dns_sd_service_number_of_entries, NULL, BBFDM_BOTH},
{"AdvertiseNumberOfEntries", &DMREAD, DMT_UNINT, get_dns_sd_advertise_number_of_entries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.DNS.SD.Service.{i}. *** */
DMOBJ tDNSSDServiceObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"TextRecord", &DMREAD, NULL, NULL, NULL, browseDNSSDServiceTextRecordInst, NULL, NULL, NULL, tDNSSDServiceTextRecordParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tDNSSDServiceParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"InstanceName", &DMREAD, DMT_STRING, get_dns_sd_service_instance_name, NULL, BBFDM_BOTH},
{"ApplicationProtocol", &DMREAD, DMT_STRING, get_dns_sd_service_application_protocol, NULL, BBFDM_BOTH},
{"TransportProtocol", &DMREAD, DMT_STRING, get_dns_sd_service_transport_protocol, NULL, BBFDM_BOTH},
{"Domain", &DMREAD, DMT_STRING, get_dns_sd_service_domain, NULL, BBFDM_BOTH},
{"Port", &DMREAD, DMT_UNINT, get_dns_sd_service_port, NULL, BBFDM_BOTH},
{"Target", &DMREAD, DMT_STRING, get_dns_sd_service_target, NULL, BBFDM_BOTH},
{"Status", &DMREAD, DMT_STRING, get_dns_sd_service_status, NULL, BBFDM_BOTH},
{"LastUpdate", &DMREAD, DMT_TIME, get_dns_sd_service_last_update, NULL, BBFDM_BOTH},
{"Host", &DMREAD, DMT_STRING, get_dns_sd_service_host, NULL, BBFDM_BOTH, DM_FLAG_REFERENCE},
{"TimeToLive", &DMREAD, DMT_UNINT, get_dns_sd_service_time_to_live, NULL, BBFDM_BOTH},
{"Priority", &DMREAD, DMT_UNINT, get_dns_sd_service_priority, NULL, BBFDM_BOTH},
{"Weight", &DMREAD, DMT_UNINT, get_dns_sd_service_weight, NULL, BBFDM_BOTH},
{"TextRecordNumberOfEntries", &DMREAD, DMT_UNINT, get_dns_sd_service_text_record_number_of_entries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.DNS.SD.Service.{i}.TextRecord.{i}. *** */
DMLEAF tDNSSDServiceTextRecordParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Key", &DMREAD, DMT_STRING, get_dns_sd_service_text_record_key, NULL, BBFDM_BOTH},
{"Value", &DMREAD, DMT_STRING, get_dns_sd_service_text_record_value, NULL, BBFDM_BOTH},
{0}
};

DM_MAP_OBJ tDynamicObj[] = {
/* parentobj, nextobject, parameter */
{"Device.DNS.", tDNSPluginObj, NULL},
{0}
};
